/**
 * Network Object:
 *
 * This object handles all incomming networking socket connections. It
 * stores in buffers data as it arrives which it will reveal as the
 * information is requested (removing the chars from the buffer as once
 * they have been asked for). This object talks to the User Object to assign
 * a new user a socket to which it will always be referenced with.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "../common/define.h"
#include "../common/server.h"
#include "../common/user.h"
#include "../common/modutil.h"
#include "../common/netutil.h"
#include "userdrop.h"

#ifdef WIN32
#include <windows.h>
#include <winsock.h>
#else
#define close_socket close
#include <arpa/inet.h>
#include <netdb.h>
#include <netinet/in.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <sys/time.h>
#include <unistd.h>			/* write, read, close */
/*
#include <netinet/in.h>				haven't needed this yet?
*/
#endif /* WIN32 */

/**
 * Start listening on the specified port
 */
int network_startup(struct server_info_hold *server_info, int port) {
	struct sockaddr_in serv_addr;

	printf("Initializing port %d\n", port);

#ifdef WIN32
	WSADATA wsaData;

	if( WSAStartup(0x0101, &wsaData) ) {
		printf("Windows socket initialization failed.\n");
	}
#endif /* WIN32 */

	/**
	 * Open a TCP socket (an Internet stream socket)
	 */
	if((server_info->answer_socket = socket(AF_INET, SOCK_STREAM, 0)) < 0) {
		printf("server: can't open stream socket\n");
		return 0;
	}

	/**
	 * Bind out local address so that the client can send to us
	 */
	memset((char*)&serv_addr, 0, sizeof(serv_addr));
	serv_addr.sin_family = AF_INET;
	serv_addr.sin_addr.s_addr = htonl(INADDR_ANY);
	serv_addr.sin_port = htons(port);

	if(bind(server_info->answer_socket,
			(struct sockaddr*) &serv_addr, sizeof(serv_addr)) < 0) {
		printf("server: can't bind local address\n");
		return 0;
	}
	else {
		if(listen(server_info->answer_socket,
				server_info->set_max_users) != 0) {
			printf("server: listen error\n");
			return 0;
		}
	}

	return 1;
}

/* Maybe this should be something different by standards it
 * should maybe be the name of the function called when the
 * object is dealloacted
 */
/* int NetworkObj::shutdown(int) { */
/* #ifdef WIN32 */
	/* Clean up socket lib */
/*	WSACleanup(); */
/* #endif */ /* WIN32 */

/*	return 1; */
/*} */

int network_update_buffers(server_info, standard_input)
struct server_info_hold *server_info;
char **standard_input;
{
	fd_set readfs;
	struct timeval tv;
	struct sockaddr_in cli_addr;
	int max_desc, new_socket;
#ifdef WIN32
	int cli_len;
#else
	unsigned int cli_len;
#endif /* WIN32 */
	int toggle, tester, loop;
	struct hostent *from_addr;
	node_user_ptr user_slider;
	node_user_ptr new_user_ptr;
	node_user_ptr fix_user_ptr;

	int chars_read;
	char buff[512];
	unsigned char netbuff[512];

	tv.tv_sec = server_info->periodic_time_seconds;
	tv.tv_usec = server_info->periodic_time_useconds;

	FD_ZERO(&readfs);
#ifndef WIN32
	FD_SET(0, &readfs);
#endif /* WIN32 */
	FD_SET(server_info->answer_socket, &readfs);
	max_desc = server_info->answer_socket;

	user_slider = server_info->user_head;
	while(user_slider) {
		FD_SET(user_slider->socket, &readfs);
		if(user_slider->socket > max_desc) {
			max_desc = user_slider->socket;
		}
		user_slider = user_slider->next_user;
	}

	select(max_desc+1, &readfs, NULL, NULL, &tv);

	/* Check for std-in */
	if(FD_ISSET(0, &readfs)) {
		/* Read characters typed into server */
		fgets(buff, 512, stdin);
		*standard_input = strdup(buff);

		return 1;
	}

	/* Add client if one has connected, the server will realize this guy */
	/* is connected after they send valid client information, until they */
	/* do they will just sit idle. */
	else if(FD_ISSET(server_info->answer_socket, &readfs)) {
		cli_len = sizeof(cli_addr);
		new_socket = accept(server_info->answer_socket, (struct sockaddr*) &cli_addr, &cli_len);
		if(new_socket < 0) {
			printf("error: server accept\n");
		}
		from_addr = gethostbyaddr((char*)&cli_addr.sin_addr,
					sizeof(cli_addr.sin_addr), AF_INET);

		/* Check for a full server, notify the client */
		/* We won't even allocate a node if this is the case, we'll write to
		   the socket with only the message that says the server is full
		   then disconnect them
		*/
		if(server_info->number_of_users_connected
				== server_info->set_max_users) {
			printf("! server is FULL, we should be ignoring this person!\n");
			printf("! but the feature to tell them it is full isn't here\n");
			printf("! we will just kill them\n");
			close_socket(new_socket);
		}
		else {
			/* Allocate a new user node for this new user */
			new_user_ptr = (node_user_ptr) calloc (1, sizeof(node_user_type));

			/* Fill in the new information about the user */

			/* the next user node is always null by default */
			new_user_ptr->next_user = NULL;

			/* Try to resolve the host IP, store numerical if not */
			if(from_addr) {
				new_user_ptr->ip_resolve = strdup(from_addr->h_name);
			}
			else {
				/* Here should build up the ip number form */
				new_user_ptr->ip_resolve =
						strdup(inet_ntoa(cli_addr.sin_addr));
			}

			/* Add the socket to the user node */
			new_user_ptr->socket = new_socket;

			/* Set the nick to NULL so it can be known first time it changes */
			new_user_ptr->nick = NULL;

			/* Find next available id number and add it to user node */
			for(loop = 0; loop < MAX_USER_TOGGLES; loop ++) {
				if(!server_info->user_id_toggles[loop]) {
					break;
				}
			}
			server_info->user_id_toggles[loop] = TRUE;
			new_user_ptr->id_number = loop;

			/* Set the number of characters in their buffer to zero */
			new_user_ptr->num_of_chars = 0;

			/* Move to the end of the node list & attach */
			user_slider = server_info->user_head;
			if(user_slider == NULL) {
				server_info->user_head = new_user_ptr;
			}
			else if(user_slider->next_user == NULL) {
				user_slider->next_user = new_user_ptr;
			}
			else {
				do {
					user_slider = user_slider->next_user;
				} while(user_slider->next_user != NULL);
				user_slider->next_user = new_user_ptr;
			}

			/* Add another user to the number of users connected */
			server_info->number_of_users_connected ++;
		}

		return 2;
	}
	else {
		/* Check active ports for characters */
		toggle = 0;

		user_slider = server_info->user_head;
		while(user_slider) {
			if(FD_ISSET(user_slider->socket, &readfs)) {
				/* This user has sent data, cancel their ping for now */
/*				pingToggle[ctr] = 0;
*/
				/* Read characters sent by client */
				chars_read = recv(user_slider->socket, netbuff, 512, 0);
				if(chars_read == -1) {
					printf("error: read, occured (client), ignoring\n");
				}
				else if(chars_read == 0) {
					/* Some modules may want to do something at this time. */
					/* We will give them the opportunity with this function */
					network_user_dropped_connection(
							server_info, user_slider->id_number);

					/* Client has dropped connection, close their socket */
					close_socket(user_slider->socket);

					/* User no longer needs an id number */
					server_info->user_id_toggles[user_slider->id_number]
							= FALSE;

					/* Now fix the linked list so we can remove this node */
					fix_user_ptr = server_info->user_head;
					if(fix_user_ptr == user_slider) {
						server_info->user_head = user_slider->next_user;
					}
					else {
						while(fix_user_ptr->next_user != user_slider) {
							fix_user_ptr = fix_user_ptr->next_user;
						}
						fix_user_ptr->next_user = user_slider->next_user;
					}
					free(user_slider->nick);
					free(user_slider->ip_resolve);
					free(user_slider);
					
					/* Subtract another user to the number of users connected */
					server_info->number_of_users_connected --;

					return 4;
				}
				else {
					/* Toggle indication that chars were in fact read */
					toggle = 1;

					/* Values greater than 7 bits (or 127) are not allowed */
					for(loop = 0; loop < chars_read; loop ++) {
						tester = netbuff[loop];
						if(tester < 0) {
printf("Here we can kill any string w/ a char > 127, maybe later.\n");
/*							netbuff[loop] = 127;
*/						}
					}
					/* Add characters to the buffer */
					if((user_slider->num_of_chars+chars_read) < MAX_BUFFER) {
						printf("adding (%d) to: (%d) ", chars_read,
								user_slider->id_number);
						mystrncpy(
								user_slider->buffer+user_slider->num_of_chars,
								netbuff, chars_read);
						user_slider->num_of_chars += chars_read;
						show_as_hex(user_slider->num_of_chars,
								user_slider->buffer);
					}
					else {
						printf("user has sent > %d chars, ignoring them",
							MAX_BUFFER);
						printf("; you might as well kick them\n");
						printf("users buffer has been flushed");
						printf("user has sent > %d chars, ignoring them",
							MAX_BUFFER);
						printf("; you might as well kick them");
						printf("users buffer has been flushed");
						user_slider->num_of_chars = 0;
						strcpy(user_slider->buffer, "");
					}
				}
			}
			user_slider = user_slider->next_user;
		}

		if(toggle) {
			return 3;
		}
	}

	return 0;
}
