/**
 * Handles storage of robot info & contains robot functions.
 *
 * Author: Travis W. Loyd
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "../common/loadall.h"
#include "addnode.h"
#include "delnode.h"
#include "frfi.h"
#include "remove.h"
#include "spawn.h"
#include "setface.h"
#include "haveface.h"
#include "getspawn.h"
#include "getworld.h"
#include "loadmod.h"
#include "getmodid.h"
#include "setspawn.h"
#include "s_robot.h"
#include "shutdown.h"
#include "virtual.h"

/* Prototypes */
void remove_nonplayers_from_robot_list(struct server_info_hold*);

/* Global variables related to module description */
unsigned int list_of_network_commands[]
		= { C_NOT_A_PLAYER, C_IS_A_PLAYER,
			C_ADMIN_SET_ARCHIVE_LOCATION, C_SPAWNING_DIRECTION,
	   		C_I_WANT_TO_SHUTDOWN };
struct initial_info_hold initial_info;
struct timing_hold list_timing_sessions[2];
unsigned int list_of_priorities_1[] = { 15000 };
unsigned int list_of_priorities_2[] = { 35000 };
char *multi_description[] = {
	"Implements a robot, the item controlled by the user.",
	"Robots are moved by players by playing program cards and then are",
	"effected by board elements.  Robots are armed with lasers which can",
	"damage other damagable elements.  Option cards can enhance robot",
	"abilities."
	};
char *list_of_styles[] = { "damage", "move", "user_dropped_their_connection" };
char author_of_module[] = { "Travis W. Loyd" };
char version_of_module[] = { "0.01" };

/* Global variables related to module function */
struct node_robot_type *g_robot_head;
unsigned int g_number_of_robots;
int g_next_program_card;
int initial_spawn_request;
long initial_archive_location_x;
long initial_archive_location_y;

/* Initialization function called first right after module is loaded */
void initialize(struct server_info_hold *server_info) {
	g_number_of_robots = 0;
	g_next_program_card = 0;

	/* if no initial archive location is sent the default is 0,0 */
	initial_archive_location_x = 0;
	initial_archive_location_y = 0;
}

/* Initial query by game engine to get module information, called second */
struct initial_info_hold* query_initial_info(void) {
	initial_info.num_of_timing_sessions = 2;
	list_timing_sessions[0].timing_session = BEFORE_PROGRAM_CARDS_ARE_DELT;
	list_timing_sessions[0].num_of_priorities = 1;
	list_timing_sessions[0].list_of_priorities = list_of_priorities_2;
	list_timing_sessions[1].timing_session = END_OF_TURN;
	list_timing_sessions[1].num_of_priorities = 1;
	list_timing_sessions[1].list_of_priorities = list_of_priorities_1;
	initial_info.list_timing_sessions = list_timing_sessions;

	/* Single line descriptions are required */
	initial_info.one_line_description = strdup(
		"Implements a robot, the item controlled by the user.");

	/* Multi-line descriptions are not required */
	initial_info.num_of_lines_in_multi_description = 5;
	initial_info.multi_description = multi_description;

	/* Network command handling is not required */
	initial_info.num_of_network_commands = 5;
	initial_info.list_of_network_commands = list_of_network_commands;

	/* A style is not required */
	initial_info.num_of_styles = 3;
	initial_info.list_of_styles = list_of_styles;

	/* Set author & version */
	initial_info.author = author_of_module;
	initial_info.version = version_of_module;

	return &initial_info;
}

/**
 * Robots are always ready except when someone needs to send in the facing
 * direction.
 */
int is_ready(server_info)
struct server_info_hold *server_info;
{
	node_robot_ptr robot_slider;
	int value;
	int count, ptr;
	char tosend[512];

    value = all_robots_have_facing_direction(g_robot_head);

	/* This check is here just for the begining of the game and as so */
	/* is kind of a waste of cpu but come on, it won't really show up */
	/* and accomplishes the job. */
	if(value) {
		if(!server_info->set_game_in_progress) {
			/* If there are no timing sessions this function shouldn't */
			/* be called. This will happen when a game ends... kind of */
			/* a mis-fire. */
			if(!server_info->number_of_timing_sessions) {
				return TRUE;
			}

			/* we know this is the first time at the begining of the */
			/* game after all players have sent in their initial facing */
			/* direction... this is the only time when this command is */
			/* issued */

			server_info->set_game_in_progress = TRUE;

			/* send the number of players, then the player id & facing */
			/* direction of each player */
			tosend[0] = S_SPAWN_ALL_PLAYERS;

			ptr = 1;
			count = 0;
			robot_slider = g_robot_head;
			while(robot_slider) {
				robot_slider->spawn_me = FALSE;
				tosend[++ptr] = robot_slider->id_number;
				tosend[++ptr] = robot_slider->facing_direction;
				count ++;

				robot_slider = robot_slider->next_robot;
			}
			tosend[1] = count;
			network_say_to_all(server_info, tosend, ptr+1);

			/* tell client module to initially display all the robots */
			tosend[0] = S_ACTION;
			tosend[1] = get_module_id_from_name(server_info, "s_robot");
			tosend[2] = -1;
			tosend[3] = tosend[1];
			tosend[4] = -1;
			tosend[5] = 2;
			network_say_to_all(server_info, tosend, 6);
		}
		else {
			count = 0;
			robot_slider = g_robot_head;
			while(robot_slider) {
				if(robot_slider->spawn_me) {
					count++;
				}

				robot_slider = robot_slider->next_robot;
			}

			robot_slider = g_robot_head;
			while(robot_slider) {
				if(robot_slider->spawn_me) {
					robot_slider->spawn_me = FALSE;

					tosend[0] = S_ACTION;
					tosend[1] = get_module_id_from_name(server_info, "s_robot");
					tosend[2] = -1;
					tosend[3] = tosend[1];
					tosend[4] = robot_slider->id_number;
					tosend[5] = 0;
					if(count > 1) {
						tosend[6] = TRUE;
					}
					else {
						tosend[6] = FALSE;
					}
					tosend[7] = robot_slider->facing_direction;

					network_say_to_all(server_info, tosend, 8);
				}

				robot_slider = robot_slider->next_robot;
			}
		}
	}

	return value;
}

/* Called once before each game */
void init_start_game(server_info)
struct server_info_hold *server_info;
{
	char tosend[3];

	printf("[ROBOT] Initializing module...\n");

	/* Remove robot nodes which are marked as non-player */
	remove_nonplayers_from_robot_list(server_info);

	/* Tell all the players where the initial archive location is. */
	tosend[0] = S_INITIAL_ARCHIVE_LOCATION;
	tosend[1] = initial_archive_location_x;
	tosend[2] = initial_archive_location_y;
	network_say_to_all(server_info, tosend, 3);

	/* The module will now ask all robots to send their initial */
	/* facing direction. */
	initial_spawn_request = TRUE;

	/* Robot's spawn location needs to be set from the global initial set- */
	/* ting stored in the game_admin module */
	set_robots_initial_spawn_location(
			server_info, g_robot_head,
			initial_archive_location_x,
			initial_archive_location_y
			);
}

void init_quit_game(server_info)
struct server_info_hold *server_info;
{
	struct node_user_type *user_slider;
	struct node_robot_type *robot_slider;
	long kill_me;

	printf("[ROBOT] QUITTING: removing robot nodes\n");
	printf("[ROBOT] removing: ");

	robot_slider = g_robot_head;
	while(robot_slider) {
		kill_me = robot_slider->id_number;
		printf("[%ld", kill_me);

		robot_slider = robot_slider->next_robot;

    	del_robot_node(
				server_info,
				&g_robot_head,
				kill_me);
		printf("(ok)] ");
	}
	printf("\n");

	/* Now add a new robot node for each user on the server */
	user_slider = server_info->user_head;
	while(user_slider) {
		add_new_robot_node_to_link_list(
				server_info, &g_robot_head, user_slider);
		user_slider = user_slider->next_user;
	}
}

/*
void action_handle_timing(server_info, timing_session, priority)
struct server_info_hold *server_info;
unsigned int timing_session;
unsigned int priority;
{
	printf("[ROBOT] Action: Timing=%d, Priority=%d\n"
		, timing_session
		, priority);
	printf("[ROBOT] Action: Revealing Program Card #%d\n"
		, g_next_program_card);

	reveal_program_cards(server_info, g_robot_head, g_next_program_card);
	move_robots(server_info, g_robot_head, g_next_program_card);
	g_next_program_card = (++g_next_program_card) % 5;
}
*/

/* -----------------------------------------------
 * Add misc. functions other modules can call here
 * -----------------------------------------------
 */

void add_new_robot_node(server_info, user_node)
struct server_info_hold *server_info;
struct node_user_type *user_node;
{
	if(add_new_robot_node_to_link_list(server_info, &g_robot_head, user_node)) {
		++g_number_of_robots;
	}
}

int is_player(server_info, id_number)
struct server_info_hold *server_info;
unsigned int id_number;
{
	struct node_robot_type *robot_node;

	robot_node = find_robot_node_from_id_number(
			server_info, g_robot_head, id_number);

	/* The robot_node will be null if there is no node. */
	/* There will be no node if this user is not a player. */

	if(robot_node) {
		return robot_node->will_play_when_game_starts;
	}
	else {
		return FALSE;
	}
}

struct node_robot_type*
get_robot_by_id(server_info, id_number)
struct server_info_hold *server_info;
unsigned int id_number;
{
	return find_robot_node_from_id_number(
			server_info, g_robot_head, id_number);
}

void remove_nonplayers_from_robot_list(server_info)
struct server_info_hold *server_info;
{
	remove_nonplayers_from_robot_link_list(server_info, &g_robot_head);
}

/* Network commands handled from here */
int network_command_handler(server_info, user_node)
struct server_info_hold *server_info;
struct node_user_type *user_node;
{
	node_robot_ptr robot_slider;

	switch(user_node->buffer[0]) {
		case C_ADMIN_SET_ARCHIVE_LOCATION:
				printf("Got initial archive location: (%d, %d)\n",
						user_node->buffer[1],
						user_node->buffer[2]);
				initial_archive_location_x = user_node->buffer[1];
				initial_archive_location_y = user_node->buffer[2];
				remove_chars_from_user_buffer(user_node, 3);
				break;
		case C_NOT_A_PLAYER:
				remove_chars_from_user_buffer(user_node, 1);
				if(!server_info->number_of_timing_sessions) {
					printf(";; setting user to NOT be a player\n");
					robot_slider = find_robot_node_from_id_number(
							server_info, g_robot_head, user_node->id_number);
					robot_slider->will_play_when_game_starts = FALSE;
				}
				else {
					printf(";; GAME IN PROGRESS: player Y/N ignored\n");
				}
				break;
		case C_IS_A_PLAYER:
				remove_chars_from_user_buffer(user_node, 1);
				if(!server_info->number_of_timing_sessions) {
 					printf(";; setting user to be a player\n");
					robot_slider = find_robot_node_from_id_number(
							server_info, g_robot_head, user_node->id_number);
					robot_slider->will_play_when_game_starts = TRUE;
				}
				else {
					printf(";; GAME IN PROGRESS: player Y/N ignored\n");
				}
				break;
		case C_SPAWNING_DIRECTION:
				if(set_spawning_direction_in_robot(
						server_info, g_robot_head, user_node)) {
				}
				break;
		case C_I_WANT_TO_SHUTDOWN:
				remove_chars_from_user_buffer(user_node, 1);
				robot_slider = find_robot_node_from_id_number(
						server_info, g_robot_head, user_node->id_number);
				if(robot_slider) {
					robot_slider->shut_me_down = TRUE;
				}
				break;
	}

	return 0;
}
 
/* Handle timing & priorities */
int action_handle_timing(server_info, timing_session, priority)
	struct server_info_hold *server_info;
	unsigned int timing_session;
	unsigned int priority;
{
	printf("[ROBOT] Action: Timing=%d, Priority=%d\n"
		, timing_session
		, priority);

	switch(timing_session) {
		case BEFORE_PROGRAM_CARDS_ARE_DELT:
				/* Since we are only active during one priority we know this one
				 * is 35000.
				 */
				if(initial_spawn_request) {
					/* Robots start off with no spawning facing direction, */
					/* we need to get one */
					get_spawning_direction_from_all_robots(
							server_info, g_robot_head);

					/* We don't want this to happen again so set the toggle */
					initial_spawn_request = FALSE;
				}
				break;
		case END_OF_TURN:
				/* Since we are only active during one priority we know this one
				 * is 15000.
				 */
				spawn_dead_robots(server_info, g_robot_head);
				handle_shutdown(server_info, g_robot_head);
				toggle_virtual(server_info, g_robot_head);
				break;
	}

	return 0;
}                                                                                                            
int network_user_dropped_connection(server_info, id_number)
struct server_info_hold *server_info;
int id_number;
{
	printf("[DEBUG] (robot): We see that %d has dropped connection.\n",
							        id_number);
			 
    /* removed the robot node representing this user if there */
    /* is one */
    del_robot_node(server_info, &g_robot_head, id_number);

	/* if we have lost all our players stop the game */
	if(g_robot_head == NULL) {
		printf("[ROBOT] all players have left the game, stopping\n");
		server_info->set_game_in_progress = 2;
	}
			 
    return 0;
}                                                                                            

/**
 * robot movement occurs here, however, the network message is
 * sent by the module which called this function so this function
 * return a boolean
 *
 * this function searches for modules with the style 'stop' and
 * asks them if they will stop this potential movement
 */
int move_with_check(server_info, move_type, roboNum)
struct server_info_hold *server_info;
int move_type;
long roboNum;
{
	struct node_robot_type *robot_slider;
	int blocked;
	long old_x, old_y;
	long new_x, new_y;
	int old_facing, new_facing;

	printf("[ROBOT] Moving (with check) type %d\n", move_type);

	/**
	 * find robot node to move
	 */
	robot_slider = g_robot_head;
	while(robot_slider) {
		if(robot_slider->id_number == roboNum) {
			break;
		}
		robot_slider = robot_slider->next_robot;
	}
	if(!robot_slider) {
		printf("[ROBOT] This point should never be reached.\n");
		return TRUE;
	}

	if(robot_slider->damage == 10) {
		return 999;
	}

	/**
	 * store values before movement
	 */
	old_x = robot_slider->current_location_xy.x;
	old_y = robot_slider->current_location_xy.y;
	old_facing = robot_slider->facing_direction;
	new_x = old_x;
	new_y = old_y;
	new_facing = old_facing;

	/**
	 * generate the potential move
	 */
	switch(move_type) {
		case U_TURN:
			new_facing = (new_facing + 1) % 4;
			new_facing = (new_facing + 1) % 4;
			break;
		case ROTATE_LEFT:
			new_facing = (new_facing + 1) % 4;
			new_facing = (new_facing + 1) % 4;
			new_facing = (new_facing + 1) % 4;
			break;
		case ROTATE_RIGHT:
			new_facing = (new_facing + 1) % 4;
			break;
		case BACK_UP:
			switch(old_facing) {
				case 0: ++new_y; break;
				case 1: --new_x; break;
				case 2: --new_y; break;
				case 3: ++new_x; break;
			}
			break;
		case FORWARD:
			switch(old_facing) {
				case 0: --new_y; break;
				case 1: ++new_x; break;
				case 2: ++new_y; break;
				case 3: --new_x; break;
			}
			break;
		case NORTH:
			--new_y;
			break;
		case EAST:
			++new_x;
			break;
		case SOUTH:
			++new_y;
			break;
		case WEST:
			--new_x;
			break;
	}

	/**
	 * see if the movement is allowed
	 */
//	check4stop(server_info, from_x, from_y, to_x, to_y);
	blocked = check4stop(server_info, old_x, old_y, new_x, new_y);

	/**
	 * move the robot
	 */
	if(blocked == -1) {
		robot_slider->current_location_xy.x = new_x;
		robot_slider->current_location_xy.y = new_y;
		robot_slider->facing_direction = new_facing;
	}

	/* this value is the module id or -1 if not blocked */
	return (blocked);
}

/**
 * robot movement occurs here, however, the network message is
 * sent by the module which called this function so this function
 * return a boolean
 */
int move_without_check(server_info, move_type, roboNum)
struct server_info_hold *server_info;
int move_type;
long roboNum;
{
	struct node_robot_type *robot_slider;
//	int allowed;
	long old_x, old_y;
	long new_x, new_y;
	int old_facing, new_facing;

	printf("[ROBOT] Moving (without check) type %d\n", move_type);

	/**
	 * find robot node to move
	 */
	robot_slider = g_robot_head;
	while(robot_slider) {
		if(robot_slider->id_number == roboNum) {
			break;
		}
		robot_slider = robot_slider->next_robot;
	}
	if(!robot_slider) {
		printf("[ROBOT] This point should never be reached.\n");
		return TRUE;
	}

	if(robot_slider->damage == 10) {
		return 999;
	}

	/**
	 * store values before movement
	 */
	old_x = robot_slider->current_location_xy.x;
	old_y = robot_slider->current_location_xy.y;
	old_facing = robot_slider->facing_direction;
	new_x = old_x;
	new_y = old_y;
	new_facing = old_facing;

	/**
	 * generate the potential move
	 */
	switch(move_type) {
		case U_TURN:
			new_facing = (new_facing + 1) % 4;
			new_facing = (new_facing + 1) % 4;
			break;
		case ROTATE_LEFT:
			new_facing = (new_facing + 1) % 4;
			new_facing = (new_facing + 1) % 4;
			new_facing = (new_facing + 1) % 4;
			break;
		case ROTATE_RIGHT:
			new_facing = (new_facing + 1) % 4;
			break;
		case BACK_UP:
			switch(old_facing) {
				case 0: ++new_y; break;
				case 1: --new_x; break;
				case 2: --new_y; break;
				case 3: ++new_x; break;
			}
			break;
		case FORWARD:
			switch(old_facing) {
				case 0: --new_y; break;
				case 1: ++new_x; break;
				case 2: ++new_y; break;
				case 3: --new_x; break;
			}
			break;
		case NORTH:
			--new_y;
			break;
		case EAST:
			++new_x;
			break;
		case SOUTH:
			++new_y;
			break;
		case WEST:
			--new_x;
			break;
	}

	/**
	 * see if the movement is allowed
	 */
//	blocked = check4stop(server_info, old_x, old_y, new_x, new_y);

	/**
	 * move the robot
	 */
//	if(blocked == -1) {
		robot_slider->current_location_xy.x = new_x;
		robot_slider->current_location_xy.y = new_y;
		robot_slider->facing_direction = new_facing;
//	}

//	return blocked;
	return TRUE;
}

int is_here(long x, long y, long roboNum) {
	struct node_robot_type *robot_slider;

	robot_slider = g_robot_head;
	while(robot_slider) {
		if(robot_slider->id_number == roboNum) {
			if(robot_slider->damage != 10) {
				if(robot_slider->current_location_xy.x == x) {
					if(robot_slider->current_location_xy.y == y) {
						return TRUE;
					}
				}
			}
		}

		robot_slider = robot_slider->next_robot;
	}

	return FALSE;
}

int is_shutdown(long roboNum) {
	struct node_robot_type *robot_slider;

	robot_slider = g_robot_head;
	while(robot_slider) {
		if(robot_slider->id_number == roboNum) {
			if(robot_slider->is_shutdown) {
				return TRUE;
			}
		}

		robot_slider = robot_slider->next_robot;
	}

	return FALSE;
}

long num_of_instances()
{
	struct node_robot_type *robot_slider;
	long count;

	count = 0;

	robot_slider = g_robot_head;
	while(robot_slider) {

		count++;
		robot_slider = robot_slider->next_robot;
	}

	return count;
}

int destroy(long x, long y, long roboNum) {
	struct node_robot_type *robot_slider;

	robot_slider = g_robot_head;
	while(robot_slider) {
		if(robot_slider->id_number == roboNum) {
			if(robot_slider->current_location_xy.x == x) {
				if(robot_slider->current_location_xy.y == y) {
					robot_slider->damage = 10;
					return TRUE;
				}
			}
		}

		robot_slider = robot_slider->next_robot;
	}

	return FALSE;
}

struct node_robot_type* get_robot_head() {
	return g_robot_head;
}

void set_archive_location(roboNum, x, y)
long roboNum;
long x;
long y;
{
	struct node_robot_type *robot_slider;

	robot_slider = g_robot_head;
	while(robot_slider) {
		if(robot_slider->id_number == roboNum) {
			robot_slider->archive_location_xy.x = x;
			robot_slider->archive_location_xy.y = y;
			break;
		}

		robot_slider = robot_slider->next_robot;
	}
}

int tag_checkpoint(long roboNum, long checkpoint_num) {
	struct node_robot_type *robot_slider;

	robot_slider = g_robot_head;
	while(robot_slider) {
		if(robot_slider->id_number == roboNum) {
			/* ok is this the robot's next checkpoint? */
			if(robot_slider->next_checkpoint == checkpoint_num) {		
				robot_slider->next_checkpoint++;
				return TRUE;
			}
		}

		robot_slider = robot_slider->next_robot;
	}

	return FALSE;
}
